#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <norm.h>

#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <stdio.h>
#include <stdlib.h>
#include <net/ethernet.h>
#ifdef LINUX
#define __FAVOR_BSD
#endif
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <sys/time.h>
#include <errno.h>
#include <netdb.h>

#include <norm_debug.h>
#include <norm_ip.h>
#include <norm_hash.h>
#include <norm_misc.h>
#include <norm_options.h>
#include <norm_send.h>
#include <norm_tcb.h>
#include <norm_tcp.h>
#include <norm_stats.h>

typedef struct tcb_key
{
  u_short   int_port, ext_port;
  u_int32_t int_addr, ext_addr;
}
NormTCBKey;


static NormHash *tcb_hash = NULL;

static NormTCPPacket  *norm_tcp_packet_new(u_int len, u_int32_t seq_no);
static void            norm_tcp_packet_free(void *data);
static u_int           norm_tcb_hash_key(NormHash *h, void *data);
static int             norm_tcb_cmp_key(void *k1, void *k2);


static u_int
norm_tcb_hash_key(NormHash *h, void *data)
{
  NormTCBKey *k = (NormTCBKey*)data;
  D_ENTER;

  /*
  D(NOISE_LOGLEVEL, ("Hashing %i %i %i %i --> (%i) ^ (%i) -> %i, mod %i\n", 
		     k->int_port, k->int_addr,
		     k->ext_port, k->ext_addr,
		     (k->int_port ^ k->int_addr),
		     (k->ext_port ^ k->ext_addr),
		     ((k->int_port ^ k->int_addr) ^ (k->ext_port ^ k->ext_addr)),		    
		     h->num_slots));
  */	     

  D_RETURN_(((k->int_port ^ k->int_addr) ^
	     (k->ext_port ^ k->ext_addr)) %
	    (h->num_slots));
}


static int   
norm_tcb_cmp_key(void *k1, void *k2)
{
  NormTCBKey *key1 = (NormTCBKey*)k1;
  NormTCBKey *key2 = (NormTCBKey*)k2;
  
  D_ENTER;
  
  /* Check matching cases first */

  if (((key1->int_port == key2->int_port) &&
       (key1->ext_port == key2->ext_port) &&
       (key1->int_addr == key2->int_addr) &&
       (key1->ext_addr == key2->ext_addr)) ||
      ((key1->int_port == key2->ext_port) &&
       (key1->ext_port == key2->int_port) &&
       (key1->int_addr == key2->ext_addr) &&
       (key1->ext_addr == key2->int_addr)))
    {
      D_RETURN_(0);
    }
  
  if (key1->int_port < key2->int_port)
    D_RETURN_(-1);
  if (key1->int_port > key2->int_port)
    D_RETURN_(1);
  if (key1->ext_port < key2->ext_port)
    D_RETURN_(-1);
  if (key1->ext_port > key2->ext_port)
    D_RETURN_(1);
  if (key1->int_addr < key2->int_addr)
    D_RETURN_(-1);
  if (key1->int_addr > key2->int_addr)
    D_RETURN_(1);
  if (key1->ext_addr < key2->ext_addr)
    D_RETURN_(-1);
  if (key1->ext_addr > key2->ext_addr)
    D_RETURN_(1);

  printf("SHOULD NOT GET HERE\n");
  abort();

  D_RETURN_(0);
}


static void
norm_tcb_hash_item_free(void *data)
{
  NormHashItem  *it = (NormHashItem*)data;

  D_ENTER;
  D(NOISE_LOGLEVEL, ("Freeing tcb hash item.\n"));

  free(it->key);
  norm_tcb_free(it->data);
  D_RETURN;
}


static NormTCPPacket*
norm_tcp_packet_new(u_int len, u_int32_t seq_no)
{
  NormTCPPacket *p;

  D_ENTER;
  D(NOISE_LOGLEVEL, ("New NormTCPPacket len %i, seqno %u.\n", len, seq_no));
  p = NEW(NormTCPPacket);
  p->data   = malloc(len);
  p->len    = len;
  p->seq_no = seq_no;

  D_RETURN_(p);
}


static void
norm_tcp_packet_free(void*data)
{
  NormTCPPacket *p;

  D_ENTER;
  p = (NormTCPPacket*)data;

  if (!p)
    D_RETURN;

  D(NOISE_LOGLEVEL, ("Freeing NormTCPPacket.\n"));
  if (p->data)
    free(p->data);
  free(p);

  D_RETURN;
}


void      
norm_tcb_init(void)
{
  D_ENTER;
  if (!tcb_hash)
    {
      tcb_hash = norm_hash_new(norm_tcb_hash_key,
			       norm_tcb_cmp_key,
			       norm_tcb_hash_item_free,
			       NORM_TCB_HASH_SIZE,
			       NORM_TCB_SLOT_SIZE);
    }
  D_RETURN;
}


NormTCB *
norm_tcb_new(struct ip *ih, struct tcphdr *th, Interface *nif)
{
  NormTCPPacket   *p_tcp;
  NormTCB         *tcb;
  int              len, tlen;

  D_ENTER;

  D(INFO_LOGLEVEL, ("Creating NormTCB\n"));

  tcb = NEW(NormTCB);
  bzero(tcb, sizeof(NormTCB));
  
  /*is the packet coming from an internal or external host?*/
  if (norm_misc_is_trusted_address(ih->ip_src))
    {
      /*from an internal host*/
      tcb->ext_addr                 = ntohl(ih->ip_dst.s_addr);
      tcb->ext_port                 = ntohs(th->th_dport);
      tcb->int_addr                 = ntohl(ih->ip_src.s_addr);
      tcb->int_port                 = ntohs(th->th_sport);
      tcb->int_ack_no               = ntohl(th->th_ack);
      tcb->int_win                  = ntohs(th->th_win);

      len = ntohs(ih->ip_len);
      tlen = len-(ih->ip_hl*4)-(th->th_off * 4);

      tcb->first_int_seq_no         = ntohl(th->th_seq);
      tcb->int_seq_no               = tcb->first_int_seq_no + tlen;
      tcb->seen_int                 = TRUE;

      p_tcp = norm_tcp_packet_new(tlen, ntohl(th->th_seq));
      tcb->int_packets = norm_list_prepend(tcb->int_packets, p_tcp);
      tcb->num_int_packets = 1;

      memcpy(p_tcp->data, (u_char*)ih + (ih->ip_hl*4)+(th->th_off * 4), tlen);
    }
  else
    {
      /*from an external host*/
      tcb->int_addr                 = ntohl(ih->ip_dst.s_addr);
      tcb->int_port                 = ntohs(th->th_dport);
      tcb->ext_addr                 = ntohl(ih->ip_src.s_addr);
      tcb->ext_port                 = ntohs(th->th_sport);
      tcb->ext_ack_no               = ntohl(th->th_ack);

      len = ntohs(ih->ip_len);
      tlen = len-(ih->ip_hl*4)-(th->th_off * 4);

      if (tlen > 0)
	{
	  tcb->ext_last_byte_valid = TRUE;
	  tcb->ext_last_byte = ((u_char*)th)[th->th_off * 4 + tlen - 1];
	}

      tcb->first_ext_seq_no         = ntohl(th->th_seq);
      tcb->ext_seq_no               = tcb->first_ext_seq_no + tlen;
      tcb->seen_ext                 = TRUE;

      p_tcp = norm_tcp_packet_new(tlen, ntohl(th->th_seq));
      tcb->ext_packets = norm_list_prepend(tcb->ext_packets, p_tcp);
      tcb->num_ext_packets = 1;
 
      memcpy(p_tcp->data, (u_char*)ih +(ih->ip_hl*4)+(th->th_off * 4), tlen);
    }
  
  D_RETURN_(tcb);
  nif = NULL;
}


void
norm_tcb_free(NormTCB *tcb)
{
  D_ENTER;

  if (!tcb)
    D_RETURN;

  D(INFO_LOGLEVEL, ("freeing tcb with %i int and %i ext packets\n",
		    tcb->num_int_packets, tcb->num_ext_packets));

  norm_list_free(tcb->int_packets, norm_tcp_packet_free);
  norm_list_free(tcb->ext_packets, norm_tcp_packet_free);
  free(tcb);

  D_RETURN;

  /*
  prev = NULL;
  while(cur!=NULL)
    {
      if (cur == tcb)
	{
	  if (prev == NULL)
	    *first = cur->next;
	  else
	    prev->next = cur->next;

	  norm_tcb_free_tcb(tcb);
	  D_RETURN_(0);
	}
      prev = cur;
      cur = cur->next;
    }

  D(FATAL_LOGLEVEL, ("NormTCB error in line %i of %s\n", __LINE__, __FILE__));
  abort();
  */
}


void      
norm_tcb_store(NormTCB *tcb)
{
  NormTCBKey  *key;
  
  D_ENTER;
  
  if (!tcb)
    D_RETURN;

  D(NOISE_LOGLEVEL, ("Storing TCB in hashtable\n"));
  key = NEW(NormTCBKey);
  key->ext_addr = tcb->ext_addr;
  key->ext_port = tcb->ext_port;
  key->int_addr = tcb->int_addr;
  key->int_port = tcb->int_port;
  
  norm_hash_insert(tcb_hash, key, tcb);

  D_RETURN;
}


void      
norm_tcb_flush(NormTCB *tcb)
{
  NormTCBKey key;

  D_ENTER;

  if (!tcb)
    D_RETURN;

  D(NOISE_LOGLEVEL, ("Flushing TCB.\n"));
  key.ext_addr = tcb->ext_addr;
  key.ext_port = tcb->ext_port;
  key.int_addr = tcb->int_addr;
  key.int_port = tcb->int_port;
  
  norm_hash_remove(tcb_hash, &key);

  D_RETURN;
}


NormTCB *
norm_tcb_lookup(struct ip *ih, struct tcphdr *th, Interface *nif)
{
  NormTCBKey       key;
  NormHashItem    *it;

  D_ENTER;

  /* is the packet coming from an internal or external host? */
  if (norm_misc_is_trusted_address(ih->ip_src))
    {
      /* from an internal host */
      key.ext_addr = ntohl(ih->ip_dst.s_addr);
      key.ext_port = ntohs(th->th_dport);
      key.int_addr = ntohl(ih->ip_src.s_addr);
      key.int_port = ntohs(th->th_sport);
    }
  else
    {
      /* from an external host */
      key.int_addr = ntohl(ih->ip_dst.s_addr);
      key.int_port = ntohs(th->th_dport);
      key.ext_addr = ntohl(ih->ip_src.s_addr);
      key.ext_port = ntohs(th->th_sport);
    }

  if ( (it = norm_hash_find(tcb_hash, &key)))
    {
      /* We found an existing TCB */
      D(NOISE_LOGLEVEL, ("Exising TCB.\n"));
      D_RETURN_((NormTCB*)it->data);
    }
  
  D_RETURN_(NULL);
  nif = NULL;
}


int
norm_tcb_update(NormTCB *tcb, NormPacket *p, Interface *nif)
{
  NormList        *l, *lold;
  NormTCPPacket   *p_tcp;
  u_int32_t        seqno, ackno;
  int              overlap_len, overlap_start;
  int              len, payload_len;
  struct ip       *ih;
  struct tcphdr   *th;

  D_ENTER;

  ih = (struct ip*)p->network_data;
  th = (struct tcphdr*)p->transport_data;

  /*
  if (!norm_misc_is_trusted_address(ih->ip_src))
    {
      if (norm_tcb_check_window(tcb, p, nif) < 0)
	D_RETURN_(-1);
    }
  */
  seqno = ntohl(th->th_seq);
  ackno = ntohl(th->th_ack);
  len = ntohs(ih->ip_len);
  payload_len = len - (ih->ip_hl*4) - (th->th_off * 4);

  /*is the packet coming from an internal or external host?*/
  if (norm_misc_is_trusted_address(ih->ip_src))
    {
      /*from an internal host*/

      if (tcb->first_int_seq_no == 0)
	tcb->first_int_seq_no = seqno;

      tcb->int_win     = ntohs(th->th_win);
      tcb->int_ack_no  = ntohl(th->th_ack);
      tcb->int_seq_no  = ntohl(th->th_seq);
      tcb->seen_int    = TRUE;
      l = tcb->int_packets;
    }
  else
    {
      /*from an external host*/

      if (tcb->first_ext_seq_no == 0)
	tcb->first_ext_seq_no = seqno;

      tcb->ext_ack_no  = ntohl(th->th_ack);
      tcb->ext_seq_no  = ntohl(th->th_seq);
      tcb->seen_ext    = TRUE;
      l = tcb->ext_packets;

      if (payload_len > 0)
	{
	  tcb->ext_last_byte_valid = TRUE;
	  tcb->ext_last_byte = ((u_char*)th)[th->th_off * 4 + payload_len - 1];
	}

    }

  /* Does the new packet conflict with any previously stored packet?*/

  p->transport_data_recalc_checksum = FALSE;

  while (l)
    {
      p_tcp = (NormTCPPacket*)l->data;

      /*Is there any overlap in sequence space?*/
      if ((p_tcp->len > 0) && (seqno >= p_tcp->seq_no) && (seqno < (p_tcp->seq_no + p_tcp->len)))
	{
	  /*There's forward overlap*/
	  if (seqno + payload_len <= p_tcp->seq_no + p_tcp->len)
	    {
	      overlap_len = payload_len;
	      overlap_start = seqno;
	    }
	  else
	    {
	      overlap_len = p_tcp->seq_no + p_tcp->len - seqno;
	      overlap_start = seqno;
	    }
	}
      else if ((p_tcp->len > 0) && (p_tcp->seq_no >= seqno) && (p_tcp->seq_no < (seqno + payload_len)))
	{
	  /*There's backward overlap*/
	  if (p_tcp->seq_no + p_tcp->len <= seqno + payload_len)
	    {
	      overlap_len = p_tcp->len;
	      overlap_start = p_tcp->seq_no;
	    }
	  else
	    {
	      overlap_len = seqno+payload_len - p_tcp->seq_no;
	      overlap_start = p_tcp->seq_no;
	    }
	}
      else
	{
	  l = l->next;
	  continue;
	}
      
      print_tcp(INFO_LOGLEVEL, ("Retransm. -- seq: %u len: %d  prev seq: %u payload_len: %d overlap: %u\n",
				seqno, payload_len, p_tcp->seq_no, p_tcp->len, overlap_len));
      
      if (memcmp(((u_char*)th) + (th->th_off * 4) + overlap_start - seqno,
		 p_tcp->data + overlap_start - p_tcp->seq_no,
		 overlap_len) == 0)
	{
	  /*retransmission is OK*/
	  print_tcp(INFO_LOGLEVEL, ("Retransmission OK\n"));
	  /* we can't return yet because one new packet might overlap two
	     stored packets*/
	}
      else
	{
	  /*inconsistent retransmission*/
	  print_tcp(NORM_LOGLEVEL, ("Inconsistent retransmission\n"));
	  memcpy(((u_char*)th) + (th->th_off * 4) + overlap_start - seqno,
		 p_tcp->data + overlap_start - p_tcp->seq_no,
		 overlap_len);

	  norm_stats_increment_tcp_norms();
	  
	  p->transport_data_recalc_checksum = TRUE;
	}
      l = l->next;
    }

  if (p->transport_data_recalc_checksum)
    {
      th->th_sum = 0;
      th->th_sum = ~norm_tcp_checksum(ih, th, payload_len);
      p->transport_data_recalc_checksum = FALSE;
    }
  
  /*Did the packet ACK any existing packets that we can now free?*/
  /*XXX this should be in the above loop, but it's temporarily here 
    for coding simplicity*/
  
  /*Check the incoming stored packets for an outgoing new packet!*/
  if (norm_misc_is_trusted_address(ih->ip_src))
    /*from an internal host*/
    l = tcb->ext_packets;
  else
    /*from an external host*/
    l = tcb->int_packets;
  
  while (l)
    {
      p_tcp = (NormTCPPacket*)l->data;

      /* print_tcp(("Ackno: %u, stored seq_no %u, len %d\n", ackno, p->seq_no, p->len)); */

      lold = l;
      l = l->next;

      if ((p_tcp->seq_no + p_tcp->len <= ackno) &&
	  (p_tcp->seq_no + p_tcp->len > ackno - MAX_WINDOW_ALLOWED))
	{
	  /*This packet has now been acked and can be freed*/
	  D(NOISE_LOGLEVEL, ("Freeing Exising packet.\n"));
	  if (norm_misc_is_trusted_address(ih->ip_src))
	    {	      
	      tcb->ext_packets = norm_list_remove(tcb->ext_packets, lold, norm_tcp_packet_free);
	      tcb->num_ext_packets--;
	    }
	  else
	    {
	      tcb->int_packets = norm_list_remove(tcb->int_packets, lold, norm_tcp_packet_free);
	      tcb->num_int_packets--;
	    } 
	}
    }
  
  /*The packet is OK - store it*/
  p_tcp = norm_tcp_packet_new(payload_len, seqno);
  memcpy(p_tcp->data, ((u_char*)ih) + (ih->ip_hl*4) + (th->th_off * 4), payload_len);

  if (norm_misc_is_trusted_address(ih->ip_src))
    {
      /*from an internal host*/
      tcb->int_packets = norm_list_prepend(tcb->int_packets, p_tcp);
      tcb->num_int_packets++;
    }
  else
    {
      /*from an external host*/
      tcb->ext_packets = norm_list_prepend(tcb->ext_packets, p_tcp);
      tcb->num_ext_packets++;
    }

  D(NOISE_LOGLEVEL, ("Update succesful -- tcb int %i, ext %i.\n",
		     tcb->num_int_packets, tcb->num_ext_packets));  
  D_RETURN_(0);
  nif = NULL;
}


int
norm_tcb_check_window(NormTCB *tcb, NormPacket *p, Interface *nif)
{
  u_int            win_lo, win_hi;
  u_int            len, payload_len;
  u_int            overlap_len = 0, overlap_start = 0;
  struct ip       *ih;
  struct tcphdr   *th;
  u_int            pkt_lo, pkt_hi;

  D_ENTER;

  ih = (struct ip*)p->network_data;
  th = (struct tcphdr*)p->transport_data;

  len = ntohs(ih->ip_len);
  payload_len = len - (ih->ip_hl*4) - (th->th_off * 4);
  pkt_lo = ntohl(th->th_seq);
  pkt_hi = pkt_lo + payload_len;

  if (tcb->int_ack_no == 0)
    win_lo = pkt_lo;
  else
    win_lo = tcb->int_ack_no;
  win_hi = win_lo + tcb->int_win;

  D(NOISE_LOGLEVEL, ("Window check: win_lo %u  win_hi %u  pkt_lo %u  pkt_hi %u\n",
		     win_lo, win_hi, pkt_lo, pkt_hi));

  /* Calculate overlap: */
  
  if (tcb->int_win != 0)
    {
      if (pkt_lo < win_lo && win_lo < pkt_hi)
	{
	  /* Partial overlap at the beginning of the window */

	  overlap_start = win_lo;
	  overlap_len   = pkt_hi - win_lo;
	  D(NOISE_LOGLEVEL, ("Partial %i-byte overlap at beginning.\n", overlap_len));
	}
      else if (win_lo <= pkt_lo && pkt_lo < win_hi)
	{
	  /* Overlap starting inside the window ... */
	  overlap_start = pkt_lo;

	  if (pkt_hi <= win_hi) /* ... and ending inside window. */
	    {
	      overlap_len = payload_len;
	    }
	  else                  /* ... and ending beyond window. */
	    {
	      overlap_len = win_hi - pkt_lo;
	      D(NOISE_LOGLEVEL, ("Partial %i-byte overlap at end.\n", overlap_len));
	    }
	}
      else
	{
	  D(NOISE_LOGLEVEL, ("Outside window.\n"));
	}
    }
  
  if (overlap_start)
    {
      if (overlap_len != payload_len)
	{
	  /* There is an overlap -- now adjust segment: */
	  u_char *payload = (u_char*)th + (th->th_off * 4);
	  memmove(payload, payload + (overlap_start - pkt_lo), overlap_len);
	  
	  th->th_seq = htonl(overlap_start);
	  ih->ip_len = htons(len - (payload_len - overlap_len));
	  p->ph->caplen -= (payload_len - overlap_len);
	  p->ph->len    -= (payload_len - overlap_len);

	  ih->ip_sum = 0;
	  ih->ip_sum = norm_ip_in_cksum((u_short*) ih, ih->ip_hl*4, 0);
	  th->th_sum = 0;
	  th->th_sum = ~norm_tcp_checksum(ih, th, overlap_len);

	  norm_stats_increment_tcp_norms();	  
	}
      D_RETURN_(0);
    }
  else
    {
      /* Okay, no overlap. Send keepalive segment, with last acked
       * byte as payload, if available ...
       */
      
      NormPacket     p_keepalive;
      struct ip     *ih_keepalive;
      struct tcphdr *th_keepalive;
      int            keepalive_len;
      
      norm_stats_increment_tcp_norms();
      bzero(&p_keepalive, sizeof(NormPacket));

      if (tcb->ext_last_byte_valid)
	keepalive_len = ih->ip_hl * 4 + th->th_off * 4 + 1;
      else
	keepalive_len = ih->ip_hl * 4 + th->th_off * 4;

      norm_packet_create(&p_keepalive, keepalive_len, nif);

      ih_keepalive = (struct ip *)p_keepalive.network_data;
      memcpy(ih_keepalive, ih, keepalive_len);
      
      ih_keepalive->ip_len = htons(keepalive_len);
      th_keepalive = (struct tcphdr *)(((char*)ih_keepalive) + ((u_int)th - (u_int)ih));
      
      if ((tcb->ext_last_byte_valid))
	{
	  ((u_char*)ih_keepalive)[keepalive_len - 1] = tcb->ext_last_byte;
	  th_keepalive->th_seq = htonl(tcb->int_ack_no - 1);
	}

      th_keepalive->th_flags = TH_ACK;
      ih_keepalive->ip_sum = 0;
      ih_keepalive->ip_sum = norm_ip_in_cksum((u_short*) ih_keepalive, ih_keepalive->ip_hl*4, 0);
      th_keepalive->th_sum = 0;

      if ((tcb->ext_last_byte_valid))
	th_keepalive->th_sum = ~norm_tcp_checksum(ih_keepalive, th_keepalive, 1);
      else
	th_keepalive->th_sum = ~norm_tcp_checksum(ih_keepalive, th_keepalive, 0);

      norm_packet_send(&p_keepalive, nif);
      norm_empty_packet(&p_keepalive);
    }

  if (tcb->int_win == 0)
    {
      D_RETURN_(0);
    }

  D_RETURN_(-1);
}
